#
#   onnx2mlx
#   Copyright © 2026 NatML Inc. All Rights Reserved.
#

from math import prod
import mlx.core as mx

from ..context import ConvertContext
from .._utils import onnx_dtype_to_mlx, onnx_tensor_to_mlx
from . import register


@register("Constant")
def constant(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext
) -> list[mx.array]:
    if "value" in attrs:
        return [onnx_tensor_to_mlx(attrs["value"])]
    if "value_float" in attrs:
        return [mx.array(attrs["value_float"])]
    if "value_int" in attrs:
        return [mx.array(attrs["value_int"], dtype=mx.int64)]
    if "value_floats" in attrs:
        return [mx.array(list(attrs["value_floats"]))]
    if "value_ints" in attrs:
        return [mx.array(list(attrs["value_ints"]), dtype=mx.int64)]
    raise ValueError("Constant node has no value attribute")

@register("Cast")
def cast(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext
) -> list[mx.array]:
    to = attrs.get("to")
    dtype = onnx_dtype_to_mlx(to)
    if dtype == mx.float64 and ctx.float64_mode == "emulate_fp32":
        dtype = mx.float32
    return [inputs[0].astype(dtype)]

@register("CastLike")
def cast_like(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext
) -> list[mx.array]:
    return [inputs[0].astype(inputs[1].dtype)]

@register("Identity")
def identity(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext
) -> list[mx.array]:
    return [inputs[0]]

@register("Dropout")
def dropout(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext
) -> list[mx.array]:
    outputs = [inputs[0]]
    if len(inputs) > 1:
        outputs.append(mx.ones_like(inputs[0]).astype(mx.bool_))
    return outputs

@register("Range")
def range_(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext
) -> list[mx.array]:
    start = inputs[0].item()
    limit = inputs[1].item()
    delta = inputs[2].item()
    return [mx.arange(start, limit, delta)]

@register("NonZero")
def non_zero(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext
) -> list[mx.array]:
    raise NotImplementedError("NonZero is not supported in MLX")

@register("RandomUniformLike")
def random_uniform_like(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext
) -> list[mx.array]:
    x = inputs[0]
    low = float(attrs.get("low", 0.0))
    high = float(attrs.get("high", 1.0))
    dtype = onnx_dtype_to_mlx(attrs["dtype"]) if "dtype" in attrs else x.dtype
    return [mx.random.uniform(low=low, high=high, shape=x.shape, dtype=dtype)]

@register("RandomNormalLike")
def random_normal_like(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext
) -> list[mx.array]:
    x = inputs[0]
    mean = float(attrs.get("mean", 0.0))
    scale = float(attrs.get("scale", 1.0))
    dtype = onnx_dtype_to_mlx(attrs["dtype"]) if "dtype" in attrs else x.dtype
    return [mx.add(mx.multiply(mx.random.normal(shape=x.shape, dtype=dtype), scale), mean)]

@register("ScatterElements")
def scatter_elements(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext
) -> list[mx.array]:
    data = inputs[0]
    indices = inputs[1]
    updates = inputs[2]
    axis = int(attrs.get("axis", 0))
    reduction = str(attrs.get("reduction", "none"))
    if axis < 0:
        axis += data.ndim
    N = data.shape[axis]
    K = indices.shape[axis]
    perm = [*range(data.ndim)]
    perm.append(perm.pop(axis))
    inv_perm = [0] * data.ndim
    for i, p in enumerate(perm):
        inv_perm[p] = i
    data_t = mx.transpose(data, perm)
    indices_t = mx.transpose(indices, perm)
    updates_t = mx.transpose(updates, perm).astype(data.dtype)
    batch_shape = data_t.shape[:-1]
    M = prod(batch_shape)
    data_flat = mx.reshape(data_t, (M, N))
    indices_flat = mx.reshape(indices_t, (M, K))
    indices_flat = mx.where(indices_flat < 0, indices_flat + N, indices_flat)
    updates_flat = mx.reshape(updates_t, (M, K))
    current = mx.take_along_axis(data_flat, indices_flat.astype(mx.int32), axis=-1)
    match reduction:
        case "none":
            delta = updates_flat - current
        case "add":
            delta = updates_flat
        case "mul":
            delta = current * updates_flat - current
        case _:
            raise NotImplementedError(f"ScatterElements reduction '{reduction}'")
    compute_dtype = mx.float32 if mx.issubdtype(data.dtype, mx.integer) else data.dtype
    indicator = mx.equal(
        indices_flat[:, None, :],
        mx.arange(N)[None, :, None]
    ).astype(compute_dtype)
    scattered = mx.sum(indicator * delta[:, None, :].astype(compute_dtype), axis=-1)
    result_flat = data_flat + scattered.astype(data.dtype)
    result_t = mx.reshape(result_flat, [*batch_shape, N])
    return [mx.transpose(result_t, inv_perm)]

@register("ScatterND")
def scatter_nd(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext
) -> list[mx.array]:
    data = inputs[0]
    indices = inputs[1]
    updates = inputs[2]
    reduction = attrs.get("reduction", "none")
    index_depth = indices.shape[-1]
    indexed_shape = data.shape[:index_depth]
    slice_shape = data.shape[index_depth:]
    strides = [prod(data.shape[i + 1:index_depth]) for i in range(index_depth)]
    strides_arr = mx.array(strides, dtype=mx.int32)
    flat_indices = mx.sum(indices.astype(mx.int32) * strides_arr, axis=-1)
    flat_indices = mx.reshape(flat_indices, (-1,))
    num_slices = 1
    for s in indexed_shape:
        num_slices *= s
    slice_size = 1
    for s in slice_shape:
        slice_size *= s
    data_flat = mx.reshape(data, (num_slices, slice_size))
    updates_flat = mx.reshape(updates.astype(data.dtype), (-1, slice_size))
    match reduction:
        case "none":
            data_at_targets = mx.take(data_flat, flat_indices, axis=0)
            scatter_values = mx.subtract(updates_flat, data_at_targets)
        case "add":
            scatter_values = updates_flat
        case _:
            raise NotImplementedError(f"ScatterND reduction '{reduction}' is not supported")
    arange = mx.arange(num_slices)
    indicator = mx.equal(arange[None, :], flat_indices[:, None]).astype(mx.float32)
    compute_vals = scatter_values.astype(mx.float32) if mx.issubdtype(data.dtype, mx.integer) else scatter_values
    scattered = indicator.T @ compute_vals
    if mx.issubdtype(data.dtype, mx.integer):
        scattered = scattered.astype(data.dtype)
    result_flat = mx.add(data_flat, scattered)
    return [mx.reshape(result_flat, data.shape)]



#
#   onnx2mlx
#   Copyright © 2026 NatML Inc. All Rights Reserved.
#

import mlx.core as mx
from . import register

@register("SequenceEmpty")
def sequence_empty(inputs, attrs, ctx):
    return [[]]

@register("SplitToSequence")
def split_to_sequence(inputs, attrs, ctx):
    x = inputs[0]
    split = inputs[1] if len(inputs) > 1 and inputs[1] is not None else None
    axis = int(attrs.get("axis", 0))
    keepdims = bool(attrs.get("keepdims", 1))
    if split is None:
        chunks = [mx.take(x, mx.array([i]), axis=axis) if keepdims
                  else mx.squeeze(mx.take(x, mx.array([i]), axis=axis), axis=axis)
                  for i in range(x.shape[axis])]
    else:
        split_sizes = split.tolist()
        if isinstance(split_sizes, (int, float)):
            split_sizes = [int(split_sizes)] * (x.shape[axis] // int(split_sizes))
        chunks = []
        start = 0
        for s in split_sizes:
            s = int(s)
            idx = mx.arange(start, start + s)
            chunk = mx.take(x, idx, axis=axis)
            if not keepdims and s == 1:
                chunk = mx.squeeze(chunk, axis=axis)
            chunks.append(chunk)
            start += s
    return [chunks]

@register("ConcatFromSequence")
def concat_from_sequence(inputs, attrs, ctx):
    seq = inputs[0]
    axis = int(attrs.get("axis", 0))
    new_axis = bool(attrs.get("new_axis", 0))
    if new_axis:
        return [mx.stack(seq, axis=axis)]
    return [mx.concatenate(seq, axis=axis)]

@register("SequenceAt")
def sequence_at(inputs, attrs, ctx):
    seq = inputs[0]
    idx = inputs[1].item()
    return [seq[int(idx)]]

@register("SequenceInsert")
def sequence_insert(inputs, attrs, ctx):
    seq = list(inputs[0])
    tensor = inputs[1]
    if len(inputs) > 2 and inputs[2] is not None:
        pos = inputs[2].item()
        seq.insert(int(pos), tensor)
    else:
        seq.append(tensor)
    return [seq]


#
#   onnx2mlx
#   Copyright © 2026 NatML Inc. All Rights Reserved.
#

from dataclasses import dataclass
from typing import Literal

Float64Mode = Literal["raise", "use_fp32"]

@dataclass(frozen=True)
class ConvertContext:
    """
    Runtime context threaded through every op handler.
    """
    float64_mode: Float64Mode = "raise"


#
#   onnx2mlx
#   Copyright © 2026 NatML Inc. All Rights Reserved.
#

import mlx.core as mx
import mlx.nn as nn

from ..context import ConvertContext
from . import register

@register("Resize")
def resize(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    x = inputs[0]
    # inputs: X, roi, scales, [sizes]
    scales = inputs[2] if len(inputs) > 2 and inputs[2] is not None else None
    sizes = inputs[3] if len(inputs) > 3 and inputs[3] is not None else None
    mode = attrs.get("mode", "nearest")
    coord_transform = attrs.get("coordinate_transformation_mode", "half_pixel")
    align_corners = coord_transform == "align_corners"
    ndim = x.ndim - 2
    if sizes is not None:
        sizes_list = sizes.tolist()
        spatial_out = sizes_list[2:]
        spatial_in = list(x.shape[2:])
        scale_factors = tuple(spatial_out[i] / spatial_in[i] for i in range(ndim))
    elif scales is not None:
        scales_list = scales.tolist()
        scale_factors = tuple(scales_list[2:])
    else:
        return [x]
    mlx_mode = _onnx_resize_mode(mode)
    # ONNX: NCHW -> MLX Upsample expects NHWC
    if ndim == 1:
        x = mx.transpose(x, (0, 2, 1))
    elif ndim == 2:
        x = mx.transpose(x, (0, 2, 3, 1))
    elif ndim == 3:
        x = mx.transpose(x, (0, 2, 3, 4, 1))
    scale_factors = (
        scale_factors[0]
        if len(set(scale_factors)) == 1
        else scale_factors
    )
    up = nn.Upsample(
        scale_factor=scale_factors,
        mode=mlx_mode,
        align_corners=align_corners
    )
    y = up(x)
    if ndim == 1:
        y = mx.transpose(y, (0, 2, 1))
    elif ndim == 2:
        y = mx.transpose(y, (0, 3, 1, 2))
    elif ndim == 3:
        y = mx.transpose(y, (0, 4, 1, 2, 3))
    return [y]

@register("GridSample")
def grid_sample(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    x = inputs[0]        # (N, C, H_in, W_in) -- NCHW
    grid = inputs[1]     # (N, H_out, W_out, 2)
    mode = attrs.get("mode", "bilinear")
    padding_mode = attrs.get("padding_mode", "zeros")
    align_corners = bool(attrs.get("align_corners", 0))
    if mode in ("bilinear", "linear") and padding_mode == "zeros" and not align_corners:
        x_nhwc = mx.transpose(x, (0, 2, 3, 1))
        y_nhwc = (
            _grid_sample_metal(x_nhwc, grid)
            if mx.metal.is_available()
            else _grid_sample_bilinear(x_nhwc, grid)
        )
        return [mx.transpose(y_nhwc, (0, 3, 1, 2))]
    raise NotImplementedError(
        f"GridSample with mode='{mode}', padding_mode='{padding_mode}', "
        f"align_corners={align_corners} is not supported"
    )

def _onnx_resize_mode(mode: str) -> str:
    match mode:
        case "nearest":     return "nearest"
        case "linear":      return "linear"
        case "bilinear":    return "linear"
        case "cubic":       return "cubic"
        case "bicubic":     return "cubic"
        case _:             raise NotImplementedError(f"Resize mode '{mode}' is not supported")

def _grid_sample_metal(x: mx.array, grid: mx.array):
    """
    Bilinear grid sample using a custom Metal kernel.

    Parameters:
        x (mx.array): Input tensor in NHWC layout `(B, H, W, C)`.
        grid (mx.array): Sampling grid `(B, gH, gW, 2)` with coords in `[-1, 1]`.
    """
    B, _, _, C = x.shape
    _, gN, gM, _ = grid.shape
    out_shape = (B, gN, gM, C)
    kernel = mx.fast.metal_kernel(
        name="grid_sample",
        input_names=["x", "grid"],
        output_names=["out"],
        source=_GRID_SAMPLE_SOURCE,
    )
    num_elements = 1
    for s in out_shape:
        num_elements *= s
    outputs = kernel(
        inputs=[x, grid],
        output_shapes=[out_shape],
        output_dtypes=[x.dtype],
        grid=(num_elements, 1, 1),
        threadgroup=(256, 1, 1),
        template=[("T", x.dtype)],
    )
    return outputs[0]

def _grid_sample_bilinear(x: mx.array, grid: mx.array):
    """
    Bilinear grid sample using pure MLX ops.

    Parameters:
        x (mx.array): Input tensor in NHWC layout `(B, H, W, C)`.
        grid (mx.array): Sampling grid `(B, gH, gW, 2)` with coords in `[-1, 1]`.
    """
    B, H, W, C = x.shape
    _, gH, gW, _ = grid.shape
    ix = ((grid[..., 0] + 1) * W - 1) / 2
    iy = ((grid[..., 1] + 1) * H - 1) / 2
    ix0 = mx.floor(ix).astype(mx.int32)
    iy0 = mx.floor(iy).astype(mx.int32)
    ix1 = ix0 + 1
    iy1 = iy0 + 1
    fx = (ix - ix0.astype(x.dtype))[..., None]
    fy = (iy - iy0.astype(x.dtype))[..., None]
    wa = (1 - fx) * (1 - fy)
    wb = fx * (1 - fy)
    wc = (1 - fx) * fy
    wd = fx * fy
    x_flat = mx.reshape(x, (B * H * W, C))
    offsets = (mx.arange(B) * H * W).reshape(B, 1, 1)
    def _gather(iy, ix):
        mask = (iy >= 0) & (iy < H) & (ix >= 0) & (ix < W)
        idx = mx.clip(iy, 0, H - 1) * W + mx.clip(ix, 0, W - 1) + offsets
        vals = x_flat[idx.reshape(-1)].reshape(B, gH, gW, C)
        return mx.where(mask[..., None], vals, 0)
    return wa * _gather(iy0, ix0) + wb * _gather(iy0, ix1) + wc * _gather(iy1, ix0) + wd * _gather(iy1, ix1)

_GRID_SAMPLE_SOURCE = """
uint elem = thread_position_in_grid.x;
int H = x_shape[1];
int W = x_shape[2];
int C = x_shape[3];
int w_stride = C;
int h_stride = W * w_stride;
int b_stride = H * h_stride;
int gH = grid_shape[1];
int gW = grid_shape[2];
uint grid_idx = elem / C * 2;
float ix = ((grid[grid_idx] + 1) * W - 1) / 2;
float iy = ((grid[grid_idx + 1] + 1) * H - 1) / 2;
int ix_nw = floor(ix);
int iy_nw = floor(iy);
int ix_ne = ix_nw + 1;
int iy_ne = iy_nw;
int ix_sw = ix_nw;
int iy_sw = iy_nw + 1;
int ix_se = ix_nw + 1;
int iy_se = iy_nw + 1;
T nw = (ix_se - ix)    * (iy_se - iy);
T ne = (ix    - ix_sw) * (iy_sw - iy);
T sw = (ix_ne - ix)    * (iy    - iy_ne);
T se = (ix    - ix_nw) * (iy    - iy_nw);
int batch_idx = elem / C / gH / gW * b_stride;
int channel_idx = elem % C;
int base_idx = batch_idx + channel_idx;
T I_nw = x[base_idx + iy_nw * h_stride + ix_nw * w_stride];
T I_ne = x[base_idx + iy_ne * h_stride + ix_ne * w_stride];
T I_sw = x[base_idx + iy_sw * h_stride + ix_sw * w_stride];
T I_se = x[base_idx + iy_se * h_stride + ix_se * w_stride];
I_nw = iy_nw >= 0 && iy_nw <= H - 1 && ix_nw >= 0 && ix_nw <= W - 1 ? I_nw : 0;
I_ne = iy_ne >= 0 && iy_ne <= H - 1 && ix_ne >= 0 && ix_ne <= W - 1 ? I_ne : 0;
I_sw = iy_sw >= 0 && iy_sw <= H - 1 && ix_sw >= 0 && ix_sw <= W - 1 ? I_sw : 0;
I_se = iy_se >= 0 && iy_se <= H - 1 && ix_se >= 0 && ix_se <= W - 1 ? I_se : 0;
out[elem] = nw * I_nw + ne * I_ne + sw * I_sw + se * I_se;
"""



#
#   onnx2mlx
#   Copyright © 2026 NatML Inc. All Rights Reserved.
#

import mlx.core as mx
from . import register

@register("BatchNormalization")
def batch_normalization(inputs, attrs, ctx):
    x = inputs[0]
    scale = inputs[1]
    bias = inputs[2]
    mean = inputs[3]
    var = inputs[4]
    epsilon = attrs.get("epsilon", 1e-5)
    # ONNX: x is (N, C, ...), normalize over C dimension
    # Reshape scale/bias/mean/var to broadcast: (1, C, 1, 1, ...)
    shape = [1, -1] + [1] * (x.ndim - 2)
    scale = mx.reshape(scale, shape)
    bias = mx.reshape(bias, shape)
    mean = mx.reshape(mean, shape)
    var = mx.reshape(var, shape)
    y = mx.add(mx.multiply(mx.multiply(mx.subtract(x, mean), mx.rsqrt(mx.add(var, epsilon))), scale), bias)
    return [y]

@register("LayerNormalization")
def layer_normalization(inputs, attrs, ctx):
    x = inputs[0]
    scale = inputs[1]
    bias = inputs[2] if len(inputs) > 2 and inputs[2] is not None else None
    axis = attrs.get("axis", -1)
    epsilon = attrs.get("epsilon", 1e-5)
    axes = list(range(axis if axis >= 0 else x.ndim + axis, x.ndim))
    mean = mx.mean(x, axis=axes, keepdims=True)
    var = mx.mean(mx.square(mx.subtract(x, mean)), axis=axes, keepdims=True)
    y = mx.multiply(mx.multiply(mx.subtract(x, mean), mx.rsqrt(mx.add(var, epsilon))), scale)
    if bias is not None:
        y = mx.add(y, bias)
    return [y]

@register("InstanceNormalization")
def instance_normalization(inputs, attrs, ctx):
    x = inputs[0]
    scale = inputs[1]
    bias = inputs[2]
    epsilon = attrs.get("epsilon", 1e-5)
    # Normalize over spatial dims (2, 3, ...)
    spatial_axes = list(range(2, x.ndim))
    mean = mx.mean(x, axis=spatial_axes, keepdims=True)
    var = mx.mean(mx.square(mx.subtract(x, mean)), axis=spatial_axes, keepdims=True)
    shape = [1, -1] + [1] * (x.ndim - 2)
    scale = mx.reshape(scale, shape)
    bias = mx.reshape(bias, shape)
    y = mx.add(mx.multiply(mx.multiply(mx.subtract(x, mean), mx.rsqrt(mx.add(var, epsilon))), scale), bias)
    return [y]

@register("GroupNormalization")
def group_normalization(inputs, attrs, ctx):
    x = inputs[0]
    scale = inputs[1]
    bias = inputs[2]
    epsilon = attrs.get("epsilon", 1e-5)
    num_groups = attrs.get("num_groups", 1)
    N, C = x.shape[0], x.shape[1]
    spatial = x.shape[2:]
    channels_per_group = C // num_groups
    x_grouped = mx.reshape(x, [N, num_groups, channels_per_group] + list(spatial))
    axes = list(range(2, x_grouped.ndim))
    mean = mx.mean(x_grouped, axis=axes, keepdims=True)
    var = mx.mean(mx.square(mx.subtract(x_grouped, mean)), axis=axes, keepdims=True)
    x_norm = mx.multiply(mx.subtract(x_grouped, mean), mx.rsqrt(mx.add(var, epsilon)))
    x_norm = mx.reshape(x_norm, x.shape)
    shape = [1, -1] + [1] * (x.ndim - 2)
    y = mx.add(mx.multiply(x_norm, mx.reshape(scale, shape)), mx.reshape(bias, shape))
    return [y]


#
#   onnx2mlx
#   Copyright © 2026 NatML Inc. All Rights Reserved.
#

import mlx.core as mx
from . import register

@register("Reshape")
def reshape(inputs, attrs, ctx):
    data = inputs[0]
    shape = inputs[1]
    shape_list = shape.tolist()
    allowzero = int(attrs.get("allowzero", 0))
    if not allowzero:
        shape_list = [
            data.shape[i] if s == 0 and i < data.ndim else s
            for i, s in enumerate(shape_list)
        ]
    return [mx.reshape(data, shape_list)]

@register("Transpose")
def transpose(inputs, attrs, ctx):
    perm = attrs.get("perm", None)
    if perm is not None:
        perm = list(perm)
    return [mx.transpose(inputs[0], axes=perm)]

@register("Flatten")
def flatten(inputs, attrs, ctx):
    axis = attrs.get("axis", 1)
    return [mx.flatten(inputs[0], start_axis=axis)]

@register("Squeeze")
def squeeze(inputs, attrs, ctx):
    x = inputs[0]
    if len(inputs) > 1 and inputs[1] is not None:
        axes = inputs[1].tolist()
    else:
        axes = attrs.get("axes", None)
    if axes is not None:
        if isinstance(axes, int):
            axes = [axes]
        for ax in sorted(axes, reverse=True):
            x = mx.squeeze(x, axis=ax)
        return [x]
    return [mx.squeeze(x)]

@register("Unsqueeze")
def unsqueeze(inputs, attrs, ctx):
    x = inputs[0]
    if len(inputs) > 1 and inputs[1] is not None:
        axes = sorted(inputs[1].tolist())
    else:
        axes = sorted(attrs.get("axes", []))
    for ax in axes:
        x = mx.expand_dims(x, axis=ax)
    return [x]

@register("Concat")
def concat(inputs, attrs, ctx):
    axis = attrs.get("axis", 0)
    valid = [inp for inp in inputs if inp is not None]
    return [mx.concatenate(valid, axis=axis)]

@register("Split")
def split(inputs, attrs, ctx):
    x = inputs[0]
    axis = attrs.get("axis", 0)
    num_outputs = attrs.get("num_outputs", None)
    if len(inputs) > 1 and inputs[1] is not None:
        split_sizes = inputs[1].tolist()
        indices = []
        acc = 0
        for s in split_sizes[:-1]:
            acc += s
            indices.append(acc)
        return list(mx.split(x, indices, axis=axis))
    elif num_outputs is not None:
        return list(mx.split(x, num_outputs, axis=axis))
    return [x]

@register("Slice")
def slice_(inputs, attrs, ctx):
    data = inputs[0]
    starts = [int(x) for x in inputs[1].tolist()]
    ends = [int(x) for x in inputs[2].tolist()]
    axes = [int(x) for x in inputs[3].tolist()] if len(inputs) > 3 and inputs[3] is not None else list(range(len(starts)))
    steps = [int(x) for x in inputs[4].tolist()] if len(inputs) > 4 and inputs[4] is not None else [1] * len(starts)
    slices = [slice(None)] * data.ndim
    for ax, start, end, step in zip(axes, starts, ends, steps):
        s = start if abs(start) < _INT64_MAX else None
        e = end if abs(end) < _INT64_MAX else None
        slices[ax] = slice(s, e, step)
    return [data[tuple(slices)]]

@register("Gather")
def gather(inputs, attrs, ctx):
    axis = attrs.get("axis", 0)
    return [mx.take(inputs[0], inputs[1], axis=axis)]

@register("GatherElements")
def gather_elements(inputs, attrs, ctx):
    axis = attrs.get("axis", 0)
    return [mx.take_along_axis(inputs[0], inputs[1], axis=axis)]

@register("Expand")
def expand(inputs, attrs, ctx):
    data = inputs[0]
    target = [int(s) for s in inputs[1].tolist()]
    data_shape = list(data.shape)
    # Pad shapes to same rank (prepend 1s to shorter one)
    while len(data_shape) < len(target):
        data_shape.insert(0, 1)
    while len(target) < len(data_shape):
        target.insert(0, 1)
    out_shape = [max(d, t) for d, t in zip(data_shape, target)]
    return [mx.broadcast_to(data, out_shape)]

@register("Pad")
def pad(inputs, attrs, ctx):
    data = inputs[0]
    pads_tensor = inputs[1]
    constant_value = inputs[2] if len(inputs) > 2 and inputs[2] is not None else mx.array(0.0)
    axes = inputs[3].tolist() if len(inputs) > 3 and inputs[3] is not None else None
    mode = attrs.get("mode", "constant")
    pads_list = [int(p) for p in pads_tensor.tolist()]
    n = len(pads_list) // 2
    if axes is not None:
        pad_widths = [(0, 0)] * data.ndim
        for i, ax in enumerate(axes):
            ax = int(ax)
            if ax < 0:
                ax += data.ndim
            pad_widths[ax] = (pads_list[i], pads_list[i + n])
    elif n < data.ndim:
        pad_widths = [(0, 0)] * (data.ndim - n) + [(pads_list[i], pads_list[i + n]) for i in range(n)]
    else:
        pad_widths = [(pads_list[i], pads_list[i + n]) for i in range(n)]
    if mode == "constant":
        return [mx.pad(data, pad_widths, constant_values=constant_value)]
    return [mx.pad(data, pad_widths)]

@register("Tile")
def tile(inputs, attrs, ctx):
    return [mx.tile(inputs[0], inputs[1].tolist())]

@register("Shape")
def shape(inputs, attrs, ctx):
    s = list(inputs[0].shape)
    start = int(attrs.get("start", 0))
    end = int(attrs.get("end", len(s)))
    return [mx.array(s[start:end], dtype=mx.int64)]

@register("ConstantOfShape")
def constant_of_shape(inputs, attrs, ctx):
    shape = inputs[0].tolist()
    value = attrs.get("value", None)
    if value is not None:
        from onnx import numpy_helper
        val = numpy_helper.to_array(value).item()
    else:
        val = 0.0
    return [mx.full(shape, val)]

_INT64_MAX = 2**63 - 1
_INT64_MIN = -(2**63 - 1)


#
#   onnx2mlx
#   Copyright © 2026 NatML Inc. All Rights Reserved.
#

from __future__ import annotations
import mlx.core as mx
from onnx import ModelProto
from pathlib import Path
from struct import pack, pack_into, unpack_from
from typing import Literal

from .convert import convert_to_mlx, ConvertContext

def export_to_mlx(
    model: ModelProto,
    path: str | Path,
    inputs: tuple[mx.array, ...],
    *,
    device: Literal["auto", "cpu", "gpu"]="auto",
    shapeless: bool=False,
    float64_mode: ConvertContext.Float64Mode="raise",
) -> None:
    """
    Convert an ONNX model to MLX and export it to a `.mlxfn` file.

    This combines `convert_to_mlx` conversion with `mx.export_function`, and
    optionally remaps all primitive streams to a specific device. This is
    useful when exporting on a CPU-only machine (e.g. Linux without Metal)
    but the `.mlxfn` will be loaded on a GPU-capable device.

    Parameters:
        path (str | Path): Output `.mlxfn` file path.
        model (ModelProto): ONNX model to convert and export.
        inputs (tuple): Sample `mx.array` inputs matching the model's runtime inputs.
        device (str): Target device for all primitives in the exported graph.
        shapeless (bool): Whether to export without baked-in shapes.
        float64_mode (str): How to handle float64 casts during conversion.
    """
    path = Path(path)
    # Convert and export
    mlx_fn = convert_to_mlx(model, float64_mode=float64_mode)
    def _export_fn(*args: mx.array):
        out = mlx_fn(*args)
        return [out] if isinstance(out, mx.array) else list(out)
    mx.export_function(str(path), _export_fn, *inputs, shapeless=shapeless)
    # Remap device streams
    if device != "auto":
        target = _DEVICE_TYPE_GPU if device == "gpu" else _DEVICE_TYPE_CPU
        source = _DEVICE_TYPE_CPU if device == "gpu" else _DEVICE_TYPE_GPU
        _remap_mlxfn_streams(path, source=source, target=target)

def _remap_mlxfn_streams(
    path: Path,
    *,
    source: int,
    target: int,
) -> int:
    """
    Rewrite primitive device streams in a serialized `.mlxfn` file.

    Each primitive's stream is serialized as 12 bytes right before its name:

        [int32 stream_index | int32 device_type | int32 device_index]

    where `device_type` 0 = CPU, 1 = GPU.

    We locate each primitive by searching for its serialized name pattern
    (uint64 length + ASCII bytes) then patch the `device_type` field
    8 bytes before the match.

    Parameters:
        path (Path): Path to the `.mlxfn` file to patch in-place.
        source (int): Device type value to replace (0 for CPU, 1 for GPU).
        target (int): Device type value to write.

    Returns:
        int: Number of streams patched.
    """
    data = bytearray(path.read_bytes())
    name_patterns = [
        pack("<Q", len(n)) + n.encode("ascii")
        for n in _MLX_PRIMITIVE_NAMES
    ]
    patched_offsets: set[int] = set()
    for pattern in name_patterns:
        pos = 0
        while True:
            idx = data.find(pattern, pos)
            if idx == -1:
                break
            dt_offset = idx - 8
            if dt_offset < 4:
                pos = idx + 1
                continue
            stream_index = unpack_from("<i", data, dt_offset - 4)[0]
            device_type = unpack_from("<i", data, dt_offset)[0]
            device_index = unpack_from("<i", data, dt_offset + 4)[0]
            if (
                device_type == source
                and device_index == 0
                and 0 <= stream_index < 256
                and dt_offset not in patched_offsets
            ):
                pack_into("<i", data, dt_offset, target)
                patched_offsets.add(dt_offset)
            pos = idx + len(pattern)
    path.write_bytes(data)
    return len(patched_offsets)

_MLX_PRIMITIVE_NAMES = [
    "Abs", "Add", "AddMM", "Arange", "ArcCos", "ArcCosh", "ArcSin",
    "ArcSinh", "ArcTan", "ArcTan2", "ArcTanh", "ArgPartition", "ArgReduce",
    "ArgSort", "AsType", "AsStrided", "BitwiseBinary", "BlockMaskedMM",
    "Broadcast", "BroadcastAxes", "Ceil", "Concatenate", "Conjugate",
    "Convolution", "Copy", "Cos", "Cosh", "Depends", "Divide", "DivMod",
    "DynamicSlice", "DynamicSliceUpdate", "Equal", "Erf", "ErfInv", "Exp",
    "Expm1", "ExpandDims", "FFT", "Flatten", "Floor", "Full", "Gather",
    "GatherAxis", "GatherMM", "Greater", "GreaterEqual", "Hadamard", "Imag",
    "Less", "LessEqual", "Log", "Log1p", "LogicalNot", "LogicalAnd",
    "LogicalOr", "LogAddExp", "LogSumExp", "MaskedScatter", "Matmul",
    "Maximum", "Minimum", "Multiply", "Negative", "NotEqual", "Reshape",
    "NumberOfElements", "Pad", "Partition", "Power", "QuantizedMatmul",
    "GatherQMM", "RandomBits", "Real", "Remainder", "Reduce", "Round",
    "Scan", "Scatter", "ScatterAxis", "Select", "Sigmoid", "Sign", "Sin",
    "Sinh", "Slice", "SliceUpdate", "Softmax", "Sort", "Split", "Square",
    "Squeeze", "Sqrt", "StopGradient", "Subtract", "Tan", "Tanh", "View",
    "Transpose", "Unflatten", "QRF", "SVD", "Inverse", "Cholesky", "Eig",
    "Eigh", "Quantize", "RMSNorm", "RMSNormVJP", "LayerNorm",
    "LayerNormVJP", "RoPE", "ScaledDotProductAttention", "CustomKernel",
]

_DEVICE_TYPE_CPU = 0
_DEVICE_TYPE_GPU = 1


#
#   onnx2mlx
#   Copyright © 2026 NatML Inc. All Rights Reserved.
#

import mlx.core as mx

from ..context import ConvertContext
from . import register

@register("If")
def if_(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    cond = inputs[0]
    scope = attrs.get("_scope", {})
    then_branch = attrs["then_branch"]
    else_branch = attrs["else_branch"]
    if cond.item():
        return then_branch(scope, ctx)
    return else_branch(scope, ctx)

@register("Loop")
def loop(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    max_trip = inputs[0]
    cond = inputs[1]
    body = attrs["body"]
    scope = attrs.get("_scope", {})
    carried = list(inputs[2:])
    max_trip_val = max_trip.item() if max_trip is not None else float("inf")
    cond_val = cond.item() if cond is not None else True
    num_carried = len(carried)
    num_scan = len(body._output_names) - 1 - num_carried
    scan_outputs = [[] for _ in range(num_scan)]
    for i in range(int(max_trip_val)):
        if not cond_val:
            break
        iter_val = mx.array([i], dtype=mx.int64)
        cond_arr = mx.array(cond_val, dtype=mx.bool_)
        results = body(scope, ctx, iter_val, cond_arr, *carried)
        cond_val = results[0].item() if isinstance(results[0], mx.array) else bool(results[0])
        carried = list(results[1:1 + num_carried])
        for j in range(num_scan):
            scan_outputs[j].append(results[1 + num_carried + j])
    final_scan = []
    for scan in scan_outputs:
        if len(scan) > 0 and isinstance(scan[0], list):
            final_scan.append(scan[-1] if scan else [])
        elif len(scan) > 0:
            final_scan.append(mx.concatenate(scan, axis=0) if scan else mx.array([]))
        else:
            final_scan.append([])
    return list(carried) + final_scan


#
#   onnx2mlx
#   Copyright © 2026 NatML Inc. All Rights Reserved.
#

import mlx.core as mx
import mlx.nn as nn
from . import register
from .._utils import onnx_pads_to_mlx

@register("MaxPool")
def max_pool(inputs, attrs, ctx):
    x = inputs[0]
    kernel_shape, strides, pads, auto_pad, ceil_mode = _get_pool_params(attrs)
    padding = _pool_padding(auto_pad, pads)
    ndim = x.ndim - 2
    if ndim == 1:
        x = mx.transpose(x, (0, 2, 1))
        pool = nn.MaxPool1d(kernel_size=kernel_shape[0], stride=strides[0], padding=padding)
        y = pool(x)
        y = mx.transpose(y, (0, 2, 1))
    elif ndim == 2:
        x = mx.transpose(x, (0, 2, 3, 1))
        ks = tuple(kernel_shape) if len(kernel_shape) > 1 else kernel_shape[0]
        st = tuple(strides) if len(strides) > 1 else strides[0]
        pool = nn.MaxPool2d(kernel_size=ks, stride=st, padding=padding)
        y = pool(x)
        y = mx.transpose(y, (0, 3, 1, 2))
    else:
        raise NotImplementedError(f"MaxPool with {ndim}D spatial dims is not supported")
    return [y]

@register("AveragePool")
def average_pool(inputs, attrs, ctx):
    x = inputs[0]
    kernel_shape, strides, pads, auto_pad, ceil_mode = _get_pool_params(attrs)
    padding = _pool_padding(auto_pad, pads)
    ndim = x.ndim - 2
    if ndim == 1:
        x = mx.transpose(x, (0, 2, 1))
        pool = nn.AvgPool1d(kernel_size=kernel_shape[0], stride=strides[0], padding=padding)
        y = pool(x)
        y = mx.transpose(y, (0, 2, 1))
    elif ndim == 2:
        x = mx.transpose(x, (0, 2, 3, 1))
        ks = tuple(kernel_shape) if len(kernel_shape) > 1 else kernel_shape[0]
        st = tuple(strides) if len(strides) > 1 else strides[0]
        pool = nn.AvgPool2d(kernel_size=ks, stride=st, padding=padding)
        y = pool(x)
        y = mx.transpose(y, (0, 3, 1, 2))
    else:
        raise NotImplementedError(f"AveragePool with {ndim}D spatial dims is not supported")
    return [y]

@register("GlobalAveragePool")
def global_average_pool(inputs, attrs, ctx):
    x = inputs[0]
    spatial_axes = tuple(range(2, x.ndim))
    return [mx.mean(x, axis=spatial_axes, keepdims=True)]

@register("GlobalMaxPool")
def global_max_pool(inputs, attrs, ctx):
    x = inputs[0]
    spatial_axes = tuple(range(2, x.ndim))
    return [mx.max(x, axis=spatial_axes, keepdims=True)]

def _get_pool_params(attrs):
    kernel_shape = attrs.get("kernel_shape", [1])
    strides = attrs.get("strides", [1])
    pads = attrs.get("pads", None)
    auto_pad = attrs.get("auto_pad", "NOTSET")
    ceil_mode = attrs.get("ceil_mode", 0)
    return kernel_shape, strides, pads, auto_pad, ceil_mode

def _pool_padding(auto_pad, pads):
    if auto_pad == "VALID":
        return 0
    if pads is not None:
        pairs = onnx_pads_to_mlx(pads)
        if all(p[0] == p[1] for p in pairs):
            symmetric = tuple(p[0] for p in pairs)
            return symmetric[0] if len(symmetric) == 1 else symmetric
        return pairs
    return 0


#
#   onnx2mlx
#   Copyright © 2026 NatML Inc. All Rights Reserved.
#

from onnx import TensorProto, numpy_helper
from onnx.helper import get_attribute_value
import mlx.core as mx
import numpy as np

def onnx_dtype_to_mlx(onnx_dtype: int) -> mx.Dtype:
    match onnx_dtype:
        case TensorProto.BFLOAT16:  return mx.bfloat16
        case TensorProto.FLOAT16:   return mx.float16
        case TensorProto.FLOAT:     return mx.float32
        case TensorProto.DOUBLE:    return mx.float64
        case TensorProto.INT8:      return mx.int8
        case TensorProto.INT16:     return mx.int16
        case TensorProto.INT32:     return mx.int32
        case TensorProto.INT64:     return mx.int64
        case TensorProto.UINT8:     return mx.uint8
        case TensorProto.UINT16:    return mx.uint16
        case TensorProto.UINT32:    return mx.uint32
        case TensorProto.UINT64:    return mx.uint64
        case TensorProto.BOOL:      return mx.bool_
        case _: raise ValueError(f"Unsupported ONNX data type: {onnx_dtype}")

def get_attrs(node) -> dict[str, object]:
    def _decode(v):
        return v.decode("utf-8") if isinstance(v, bytes) else v
    return { attr.name: _decode(get_attribute_value(attr)) for attr in node.attribute }

def onnx_pads_to_mlx(pads: list[int]) -> list[tuple[int, int]]:
    n = len(pads) // 2
    return [(pads[i], pads[i + n]) for i in range(n)]

def onnx_tensor_to_mlx(tensor: TensorProto) -> mx.array:
    arr = numpy_helper.to_array(tensor)
    if arr.dtype == np.float64:
        arr = arr.astype(np.float32)
    return mx.array(arr)


#
#   onnx2mlx
#   Copyright © 2026 NatML Inc. All Rights Reserved.
#

from collections import defaultdict, deque
import mlx.core as mx
from onnx import ModelProto, AttributeProto
from typing import Callable

from ._utils import get_attrs, onnx_tensor_to_mlx
from .ops import OP_REGISTRY
from .context import ConvertContext, Float64Mode

def convert_to_mlx(
    model: ModelProto,
    *,
    float64_mode: Float64Mode="raise",
) -> Callable[..., mx.array | list[mx.array]]:
    """
    Convert an ONNX model to an MLX callable.

    The returned function accepts `mx.array` inputs in the same order as the
    ONNX graph's runtime inputs (graph inputs minus initializers). It returns
    a single `mx.array` when the graph has one output, or a list otherwise.

    The callable is compatible with `mx.compile` and `mx.export_function`
    because all control flow is data-independent.

    Args:
        model: The ONNX model to convert.
        float64_mode: How to handle float64 (double) precision casts. MLX does
            not support float64 on the GPU. ``"raise"`` (default) will let the
            error propagate. ``"use_fp32"`` will downgrade float64 casts
            to float32.
    """
    ctx = ConvertContext(float64_mode=float64_mode)
    initializers, input_names, output_names, node_specs = _compile_graph(model.graph)
    def forward(*args: mx.array):
        return _run_graph(initializers, input_names, output_names, node_specs, args, ctx)
    return forward

def _compile_graph(graph) -> tuple[dict[str, mx.array], list[str], list[str], list]:
    """
    Compile an ONNX GraphProto into initializers, I/O names, and node specs.
    """
    initializers = {
        init.name: onnx_tensor_to_mlx(init)
        for init in graph.initializer
    }
    initializer_names = set(initializers.keys())
    input_names = [
        inp.name
        for inp in graph.input
        if inp.name not in initializer_names
    ]
    output_names = [out.name for out in graph.output]
    available = initializer_names | {inp.name for inp in graph.input}
    sorted_nodes = _toposort(list(graph.node), available)
    node_specs = []
    for node in sorted_nodes:
        handler = OP_REGISTRY.get(node.op_type)
        if handler is None:
            raise NotImplementedError(f"ONNX op '{node.op_type}' is not supported")
        attrs = get_attrs(node)
        for attr in node.attribute:
            if attr.type == AttributeProto.GRAPH:
                attrs[attr.name] = _compile_subgraph(attr.g)
        node_specs.append((
            list(node.input),
            list(node.output),
            handler,
            attrs,
        ))
    return initializers, input_names, output_names, node_specs

def _compile_subgraph(graph) -> Callable:
    """
    Compile a subgraph into a callable that takes a values dict.
    """
    initializers, input_names, output_names, node_specs = _compile_graph(graph)
    def run(parent_values: dict, ctx: ConvertContext, *args):
        values = dict(parent_values)
        values.update(initializers)
        for name, arg in zip(input_names, args):
            values[name] = arg
        for inp_names, out_names, handler, attrs in node_specs:
            inputs = [values.get(n) if n != "" else None for n in inp_names]
            outputs = handler(inputs, attrs, ctx)
            for name, out in zip(out_names, outputs):
                values[name] = out
        return [values[name] for name in output_names]
    run._input_names = input_names
    run._output_names = output_names
    return run

def _run_graph(initializers, input_names, output_names, node_specs, args, ctx):
    """
    Execute a compiled graph.
    """
    values = dict(initializers)
    for name, arg in zip(input_names, args):
        values[name] = arg
    for inp_names, out_names, handler, attrs in node_specs:
        inputs = [values.get(n) if n != "" else None for n in inp_names]
        outputs = handler(inputs, { **attrs, "_scope": values }, ctx)
        for name, out in zip(out_names, outputs):
            values[name] = out
    results = [values[name] for name in output_names]
    return results[0] if len(results) == 1 else results

def _toposort(nodes, available: set[str]) -> list:
    """
    Topologically sort nodes considering subgraph external references.
    """
    n = len(nodes)
    # Map output name -> node index
    producer = {}
    for i, node in enumerate(nodes):
        for out in node.output:
            if out:
                producer[out] = i
    # Build adjacency: node i depends on node j if j produces a name i needs
    in_degree = [0] * n
    dependents = defaultdict(list)
    for i, node in enumerate(nodes):
        deps = set()
        for inp in node.input:
            if inp and inp not in available and inp in producer:
                deps.add(producer[inp])
        for ext in _subgraph_external_refs(node):
            if ext not in available and ext in producer:
                deps.add(producer[ext])
        deps.discard(i)
        in_degree[i] = len(deps)
        for j in deps:
            dependents[j].append(i)
    queue = deque(i for i in range(n) if in_degree[i] == 0)
    order = []
    while queue:
        i = queue.popleft()
        order.append(i)
        for j in dependents[i]:
            in_degree[j] -= 1
            if in_degree[j] == 0:
                queue.append(j)
    if len(order) != n:
        remaining = [i for i in range(n) if in_degree[i] > 0]
        raise ValueError(f"Circular dependency among {len(remaining)} nodes")
    return [nodes[i] for i in order]

def _subgraph_external_refs(node) -> set[str]:
    """
    Collect names referenced inside subgraph attributes but not defined there.
    """
    refs = set()
    for attr in node.attribute:
        if attr.type != AttributeProto.GRAPH:
            continue
        g = attr.g
        defined = {i.name for i in g.initializer}
        defined |= {i.name for i in g.input}
        for n in g.node:
            defined.update(n.output)
        for n in g.node:
            for inp in n.input:
                if inp and inp not in defined:
                    refs.add(inp)
    return refs


#
#   onnx2mlx
#   Copyright © 2026 NatML Inc. All Rights Reserved.
#

import mlx.core as mx
import mlx.nn as nn

from ..context import ConvertContext
from . import register

@register("Relu")
def relu(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    return [nn.relu(inputs[0])]

@register("Sigmoid")
def sigmoid(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    return [mx.sigmoid(inputs[0])]

@register("Tanh")
def tanh(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    return [mx.tanh(inputs[0])]

@register("LeakyRelu")
def leaky_relu(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    alpha = attrs.get("alpha", 0.01)
    return [nn.leaky_relu(inputs[0], negative_slope=alpha)]

@register("Elu")
def elu(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    alpha = attrs.get("alpha", 1.0)
    return [nn.elu(inputs[0], alpha=alpha)]

@register("Selu")
def selu(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    return [nn.selu(inputs[0])]

@register("Gelu")
def gelu(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    approximate = attrs.get("approximate", "none")
    if approximate == "tanh":
        return [nn.gelu_approx(inputs[0])]
    return [nn.gelu(inputs[0])]

@register("Softmax")
def softmax(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    axis = attrs.get("axis", -1)
    return [mx.softmax(inputs[0], axis=axis)]

@register("LogSoftmax")
def log_softmax(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    axis = attrs.get("axis", -1)
    return [mx.log(mx.softmax(inputs[0], axis=axis))]

@register("HardSigmoid")
def hard_sigmoid(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    alpha = attrs.get("alpha", 0.2)
    beta = attrs.get("beta", 0.5)
    return [mx.clip(mx.add(mx.multiply(inputs[0], alpha), beta), 0.0, 1.0)]

@register("Hardswish")
def hardswish(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    return [nn.hardswish(inputs[0])]

@register("Clip")
def clip(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    x = inputs[0]
    a_min = inputs[1] if len(inputs) > 1 and inputs[1] is not None else None
    a_max = inputs[2] if len(inputs) > 2 and inputs[2] is not None else None
    if a_min is not None and a_max is not None:
        return [mx.clip(x, a_min, a_max)]
    elif a_min is not None:
        return [mx.maximum(x, a_min)]
    elif a_max is not None:
        return [mx.minimum(x, a_max)]
    return [x]

@register("Softplus")
def softplus(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    return [nn.softplus(inputs[0])]

@register("Softsign")
def softsign(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    return [nn.softsign(inputs[0])]

@register("Mish")
def mish(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    return [nn.mish(inputs[0])]

@register("PRelu")
def prelu(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    return [nn.prelu(inputs[0], inputs[1])]

@register("Celu")
def celu(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    alpha = attrs.get("alpha", 1.0)
    return [nn.celu(inputs[0], alpha=alpha)]


#
#   onnx2mlx
#   Copyright © 2026 NatML Inc. All Rights Reserved.
#

import mlx.core as mx

from ..context import ConvertContext
from .._utils import onnx_pads_to_mlx
from . import register

@register("Conv")
def conv(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    x = inputs[0]
    w = inputs[1]
    b = inputs[2] if len(inputs) > 2 and inputs[2] is not None else None
    return [_conv_impl(x, w, b, attrs)]

@register("ConvTranspose")
def conv_transpose(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    x = inputs[0]
    w = inputs[1]
    b = inputs[2] if len(inputs) > 2 and inputs[2] is not None else None
    strides, dilations, group, pads, auto_pad, output_padding = _get_conv_params(attrs)
    padding = _compute_padding(auto_pad, pads, x.ndim - 2)
    ndim = w.ndim - 2
    w = _conv_transpose_weight(w, group, ndim)
    if ndim == 1:
        x = mx.transpose(x, (0, 2, 1))
        y = mx.conv_transpose1d(x, w, stride=strides[0], padding=padding, dilation=dilations[0], groups=group)
        y = _apply_output_padding(y, output_padding, ndim)
        if b is not None:
            y = mx.add(y, mx.reshape(b, (1, 1, -1)))
        y = mx.transpose(y, (0, 2, 1))
    elif ndim == 2:
        x = mx.transpose(x, (0, 2, 3, 1))
        stride = tuple(strides) if len(strides) > 1 else strides[0]
        dilation = tuple(dilations) if len(dilations) > 1 else dilations[0]
        y = mx.conv_transpose2d(x, w, stride=stride, padding=padding, dilation=dilation, groups=group)
        y = _apply_output_padding(y, output_padding, ndim)
        if b is not None:
            y = mx.add(y, mx.reshape(b, (1, 1, 1, -1)))
        y = mx.transpose(y, (0, 3, 1, 2))
    elif ndim == 3:
        x = mx.transpose(x, (0, 2, 3, 4, 1))
        stride = tuple(strides)
        dilation = tuple(dilations)
        y = mx.conv_transpose3d(x, w, stride=stride, padding=padding, dilation=dilation, groups=group)
        y = _apply_output_padding(y, output_padding, ndim)
        if b is not None:
            y = mx.add(y, mx.reshape(b, (1, 1, 1, 1, -1)))
        y = mx.transpose(y, (0, 4, 1, 2, 3))
    else:
        raise NotImplementedError(f"ConvTranspose with {ndim}D spatial dims is not supported")
    return [y]

def _conv_impl(x, w, b, attrs):
    """
    Core convolution logic shared by Conv and ConvInteger.
    """
    strides, dilations, group, pads, auto_pad, _ = _get_conv_params(attrs)
    padding = _compute_padding(auto_pad, pads, x.ndim - 2)
    ndim = w.ndim - 2
    if ndim == 1:
        x = mx.transpose(x, (0, 2, 1))
        w = mx.transpose(w, (0, 2, 1))
        x, padding = _apply_asymmetric_padding(x, padding, ndim)
        y = mx.conv1d(x, w, stride=strides[0], padding=padding, dilation=dilations[0], groups=group)
        if b is not None:
            y = mx.add(y, mx.reshape(b, (1, 1, -1)))
        return mx.transpose(y, (0, 2, 1))
    elif ndim == 2:
        x = mx.transpose(x, (0, 2, 3, 1))
        w = mx.transpose(w, (0, 2, 3, 1))
        x, padding = _apply_asymmetric_padding(x, padding, ndim)
        stride = tuple(strides) if len(strides) > 1 else strides[0]
        dilation = tuple(dilations) if len(dilations) > 1 else dilations[0]
        y = mx.conv2d(x, w, stride=stride, padding=padding, dilation=dilation, groups=group)
        if b is not None:
            y = mx.add(y, mx.reshape(b, (1, 1, 1, -1)))
        return mx.transpose(y, (0, 3, 1, 2))
    elif ndim == 3:
        x = mx.transpose(x, (0, 2, 3, 4, 1))
        w = mx.transpose(w, (0, 2, 3, 4, 1))
        x, padding = _apply_asymmetric_padding(x, padding, ndim)
        stride = tuple(strides)
        dilation = tuple(dilations)
        y = mx.conv3d(x, w, stride=stride, padding=padding, dilation=dilation, groups=group)
        if b is not None:
            y = mx.add(y, mx.reshape(b, (1, 1, 1, 1, -1)))
        return mx.transpose(y, (0, 4, 1, 2, 3))
    raise NotImplementedError(f"Conv with {ndim}D spatial dims is not supported")

def _conv_transpose_weight(w, group, ndim):
    """
    Reshape ONNX ConvTranspose weight to MLX layout.

    ONNX: (C_in, C_out/groups, K...) -> MLX: (C_out, K..., C_in/groups)
    """
    c_in = w.shape[0]
    c_out_g = w.shape[1]
    spatial = w.shape[2:]
    c_in_g = c_in // group
    w = mx.reshape(w, (group, c_in_g, c_out_g, *spatial))
    # Move c_in_g to last, merge group*c_out_g
    perm = [0, 2] + list(range(3, 3 + ndim)) + [1]
    w = mx.transpose(w, perm)
    return mx.reshape(w, (group * c_out_g, *spatial, c_in_g))

def _apply_output_padding(y, output_padding, ndim):
    """
    Pad the spatial dimensions of the output by output_padding.
    """
    if output_padding is None:
        return y
    # y is in NHWC layout; spatial dims are axes 1..ndim
    pad_widths = [(0, 0)]  # batch
    for i in range(ndim):
        op = output_padding[i] if i < len(output_padding) else 0
        pad_widths.append((0, op))
    pad_widths.append((0, 0))  # channels
    return mx.pad(y, pad_widths)

def _get_conv_params(attrs):
    strides = attrs.get("strides", [1])
    dilations = attrs.get("dilations", [1])
    group = attrs.get("group", 1)
    pads = attrs.get("pads", None)
    auto_pad = attrs.get("auto_pad", "NOTSET")
    output_padding = attrs.get("output_padding", None)
    return strides, dilations, group, pads, auto_pad, output_padding

def _compute_padding(auto_pad, pads, ndim):
    """
    Compute MLX-compatible padding from ONNX padding spec.
    """
    if auto_pad == "VALID":
        return 0
    if auto_pad in ("SAME_UPPER", "SAME_LOWER"):
        return "same"
    if pads is not None:
        pairs = onnx_pads_to_mlx(pads)
        if all(p[0] == p[1] for p in pairs):
            symmetric = tuple(p[0] for p in pairs)
            return symmetric[0] if len(symmetric) == 1 else symmetric
        return pairs
    return 0

def _apply_asymmetric_padding(x, padding, ndim):
    """
    Handle asymmetric padding by pre-padding the input with zeros.

    When padding is a list of (before, after) pairs with unequal values,
    mx.conv* can't handle it directly. We pad the input manually and
    return padding=0 for the conv call.
    Input x is in channels-last layout (N, spatial..., C).
    """
    if not isinstance(padding, list):
        return x, padding
    pad_widths = [(0, 0)] + padding + [(0, 0)]
    return mx.pad(x, pad_widths), 0


#
#   onnx2mlx
#   Copyright © 2026 NatML Inc. All Rights Reserved.
#

import io
import pytest
import numpy as np
import torch
import onnx
import onnxruntime as ort
import mlx.core as mx

from onnx2mlx import onnx2mlx

def test_resnet18():
    from torchvision.models import resnet18, ResNet18_Weights
    model = resnet18(weights=ResNet18_Weights.DEFAULT)
    onnx_model, input_np = _export_to_onnx(model)
    _compare(onnx_model, input_np, rtol=1e-3, atol=1e-3)

def test_mobilenet_v2():
    from torchvision.models import mobilenet_v2, MobileNet_V2_Weights
    model = mobilenet_v2(weights=MobileNet_V2_Weights.DEFAULT)
    onnx_model, input_np = _export_to_onnx(model)
    _compare(onnx_model, input_np, rtol=1e-3, atol=1e-3)

def test_squeezenet1_1():
    from torchvision.models import squeezenet1_1, SqueezeNet1_1_Weights
    model = squeezenet1_1(weights=SqueezeNet1_1_Weights.DEFAULT)
    onnx_model, input_np = _export_to_onnx(model)
    _compare(onnx_model, input_np, rtol=1e-3, atol=1e-3)

def _run_onnxruntime(onnx_model, input_np):
    buf = io.BytesIO()
    onnx.save(onnx_model, buf)
    session = ort.InferenceSession(buf.getvalue())
    input_name = session.get_inputs()[0].name
    return session.run(None, {input_name: input_np})[0]

def _run_onnx2mlx(onnx_model, input_np):
    fn = onnx2mlx(onnx_model)
    mlx_input = mx.array(input_np)
    result = fn(mlx_input)
    mx.eval(result)
    return np.array(result)

def _compare(onnx_model, input_np, rtol=1e-4, atol=1e-4):
    expected = _run_onnxruntime(onnx_model, input_np)
    actual = _run_onnx2mlx(onnx_model, input_np)
    np.testing.assert_allclose(actual, expected, rtol=rtol, atol=atol)

def _export_to_onnx(model, input_shape=(1, 3, 224, 224), opset_version=17):
    model.eval()
    dummy = torch.randn(*input_shape)
    buf = io.BytesIO()
    torch.onnx.export(
        model,
        dummy,
        buf,
        opset_version=opset_version,
        input_names=["input"],
        output_names=["output"],
        dynamic_axes={"input": {0: "batch"}, "output": {0: "batch"}},
        dynamo=False,
    )
    buf.seek(0)
    return onnx.load(buf), dummy.numpy()


#
#   onnx2mlx
#   Copyright © 2026 NatML Inc. All Rights Reserved.
#

import mlx.core as mx

from ..context import ConvertContext
from . import register

@register("ReduceSum")
def reduce_sum(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    axes = _reduce_axes(inputs, attrs)
    keepdims = bool(attrs.get("keepdims", 1))
    return [mx.sum(inputs[0], axis=axes, keepdims=keepdims)]

@register("ReduceMean")
def reduce_mean(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    axes = _reduce_axes(inputs, attrs)
    keepdims = bool(attrs.get("keepdims", 1))
    return [mx.mean(inputs[0], axis=axes, keepdims=keepdims)]

@register("ReduceMax")
def reduce_max(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    axes = _reduce_axes(inputs, attrs)
    keepdims = bool(attrs.get("keepdims", 1))
    return [mx.max(inputs[0], axis=axes, keepdims=keepdims)]

@register("ReduceMin")
def reduce_min(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    axes = _reduce_axes(inputs, attrs)
    keepdims = bool(attrs.get("keepdims", 1))
    return [mx.min(inputs[0], axis=axes, keepdims=keepdims)]

@register("ReduceProd")
def reduce_prod(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    axes = _reduce_axes(inputs, attrs)
    keepdims = bool(attrs.get("keepdims", 1))
    return [mx.prod(inputs[0], axis=axes, keepdims=keepdims)]

@register("ArgMax")
def argmax(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    axis = attrs.get("axis", 0)
    keepdims = bool(attrs.get("keepdims", 1))
    return [mx.argmax(inputs[0], axis=axis, keepdims=keepdims)]

@register("ArgMin")
def argmin(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    axis = attrs.get("axis", 0)
    keepdims = bool(attrs.get("keepdims", 1))
    return [mx.argmin(inputs[0], axis=axis, keepdims=keepdims)]

@register("ReduceLogSumExp")
def reduce_logsumexp(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    axes = _reduce_axes(inputs, attrs)
    keepdims = bool(attrs.get("keepdims", 1))
    return [mx.logsumexp(inputs[0], axis=axes, keepdims=keepdims)]

@register("TopK")
def topk(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    x = inputs[0]
    k = inputs[1].item()
    axis = attrs.get("axis", -1)
    largest = bool(attrs.get("largest", 1))
    sorted_ = bool(attrs.get("sorted", 1))
    n = x.shape[axis]
    sort_x = mx.negative(x) if not largest else x
    # Get top-k indices via argpartition (O(n) average)
    kth = n - k
    indices = mx.argpartition(sort_x, kth=kth, axis=axis)
    indices = mx.take(indices, mx.arange(kth, n), axis=axis)
    if sorted_:
        # Sort the k candidates in descending order
        topk_vals = mx.take_along_axis(sort_x, indices, axis=axis)
        order = mx.argsort(mx.negative(topk_vals), axis=axis)
        indices = mx.take_along_axis(indices, order, axis=axis)
    values = mx.take_along_axis(x, indices, axis=axis)
    return [values, indices]

@register("CumSum")
def cumsum(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    x = inputs[0]
    axis = inputs[1].item()
    exclusive = bool(attrs.get("exclusive", 0))
    reverse = bool(attrs.get("reverse", 0))
    if reverse:
        x = mx.flip(x, axis=axis)
    if exclusive:
        x = mx.cumsum(x, axis=axis)
        # Shift forward and fill first element with 0
        slices = [slice(None)] * x.ndim
        slices[axis] = slice(0, -1)
        zero_shape = list(x.shape)
        zero_shape[axis] = 1
        x = mx.concatenate([mx.zeros(zero_shape, dtype=x.dtype), x[tuple(slices)]], axis=axis)
    else:
        x = mx.cumsum(x, axis=axis)
    if reverse:
        x = mx.flip(x, axis=axis)
    return [x]

def _reduce_axes(inputs, attrs):
    if len(inputs) > 1 and inputs[1] is not None:
        return inputs[1].tolist()
    axes = attrs.get("axes", None)
    return list(axes) if axes is not None else None


#
#   onnx2mlx
#   Copyright © 2026 NatML Inc. All Rights Reserved.
#

import io
import os
import tempfile
import numpy as np
import torch
import onnx
import onnxruntime as ort
import mlx.core as mx

from onnx2mlx import onnx2mlx

def _export_to_onnx(model, input_shape=(1, 3, 224, 224), opset_version=17):
    model.eval()
    dummy = torch.randn(*input_shape)
    buf = io.BytesIO()
    torch.onnx.export(
        model,
        dummy,
        buf,
        opset_version=opset_version,
        input_names=["input"],
        output_names=["output"],
        dynamic_axes={"input": {0: "batch"}, "output": {0: "batch"}},
        dynamo=False,
    )
    buf.seek(0)
    return onnx.load(buf), dummy.numpy()

def _run_onnxruntime(onnx_model, input_np):
    buf = io.BytesIO()
    onnx.save(onnx_model, buf)
    session = ort.InferenceSession(buf.getvalue())
    input_name = session.get_inputs()[0].name
    return session.run(None, {input_name: input_np})[0]


def test_export_resnet18():
    from torchvision.models import resnet18, ResNet18_Weights
    model = resnet18(weights=ResNet18_Weights.DEFAULT)
    onnx_model, input_np = _export_to_onnx(model)
    expected = _run_onnxruntime(onnx_model, input_np)
    fn = onnx2mlx(onnx_model)
    mlx_input = mx.array(input_np)
    with tempfile.TemporaryDirectory() as tmpdir:
        path = os.path.join(tmpdir, "resnet18.mlxfn")
        mx.export_function(path, fn, mlx_input)
        imported_fn = mx.import_function(path)
        result = imported_fn(mlx_input)
        mx.eval(result)
        np.testing.assert_allclose(np.array(result[0]), expected, rtol=1e-3, atol=1e-3)

def test_export_squeezenet1_1():
    from torchvision.models import squeezenet1_1, SqueezeNet1_1_Weights
    model = squeezenet1_1(weights=SqueezeNet1_1_Weights.DEFAULT)
    onnx_model, input_np = _export_to_onnx(model)
    expected = _run_onnxruntime(onnx_model, input_np)
    fn = onnx2mlx(onnx_model)
    mlx_input = mx.array(input_np)
    with tempfile.TemporaryDirectory() as tmpdir:
        path = os.path.join(tmpdir, "squeezenet.mlxfn")
        mx.export_function(path, fn, mlx_input)
        imported_fn = mx.import_function(path)
        result = imported_fn(mlx_input)
        mx.eval(result)
        np.testing.assert_allclose(np.array(result[0]), expected, rtol=1e-3, atol=1e-3)

def test_export_mobilenet_v2():
    from torchvision.models import mobilenet_v2, MobileNet_V2_Weights
    model = mobilenet_v2(weights=MobileNet_V2_Weights.DEFAULT)
    onnx_model, input_np = _export_to_onnx(model)
    expected = _run_onnxruntime(onnx_model, input_np)
    fn = onnx2mlx(onnx_model)
    mlx_input = mx.array(input_np)
    with tempfile.TemporaryDirectory() as tmpdir:
        path = os.path.join(tmpdir, "mobilenet_v2.mlxfn")
        mx.export_function(path, fn, mlx_input)
        imported_fn = mx.import_function(path)
        result = imported_fn(mlx_input)
        mx.eval(result)
        np.testing.assert_allclose(np.array(result[0]), expected, rtol=1e-3, atol=1e-3)

def test_export_nomic_layout_v1():
    from transformers import RTDetrForObjectDetection
    model = RTDetrForObjectDetection.from_pretrained("muna-ai/nomic-layout-v1").eval()
    torch.manual_seed(42)
    dummy = torch.randn(1, 3, 800, 800)
    buf = io.BytesIO()
    torch.onnx.export(
        model,
        dummy,
        buf,
        opset_version=17,
        input_names=["pixel_values"],
        output_names=["logits", "pred_boxes"],
        dynamo=False,
    )
    buf.seek(0)
    onnx_model = onnx.load(buf)
    input_np = dummy.numpy()
    # Run with ONNX Runtime
    buf = io.BytesIO()
    onnx.save(onnx_model, buf)
    session = ort.InferenceSession(buf.getvalue())
    ort_outputs = session.run(None, {"pixel_values": input_np})
    ort_logits = ort_outputs[0]
    ort_boxes = ort_outputs[1]
    # Convert and export
    fn = onnx2mlx(onnx_model)
    mlx_input = mx.array(input_np)
    with tempfile.TemporaryDirectory() as tmpdir:
        path = os.path.join(tmpdir, "nomic_layout.mlxfn")
        mx.export_function(path, fn, mlx_input)
        imported_fn = mx.import_function(path)
        result = imported_fn(mlx_input)
        mx.eval(result)
        mlx_logits = np.array(result[0])
        mlx_boxes = np.array(result[1])
    # Sort along query axis to handle TopK tie-breaking differences
    mlx_order = np.argsort(mlx_logits[:, :, 0], axis=1)
    ort_order = np.argsort(ort_logits[:, :, 0], axis=1)
    mlx_logits = np.take_along_axis(mlx_logits, mlx_order[:, :, None], axis=1)
    ort_logits = np.take_along_axis(ort_logits, ort_order[:, :, None], axis=1)
    mlx_boxes = np.take_along_axis(mlx_boxes, mlx_order[:, :, None], axis=1)
    ort_boxes = np.take_along_axis(ort_boxes, ort_order[:, :, None], axis=1)
    np.testing.assert_allclose(mlx_logits, ort_logits, rtol=1e-3, atol=1e-3)
    np.testing.assert_allclose(mlx_boxes, ort_boxes, rtol=1e-3, atol=1e-3)



#
#   onnx2mlx
#   Copyright © 2026 NatML Inc. All Rights Reserved.
#

import mlx.core as mx

from ..context import ConvertContext
from . import register

@register("MatMul")
def matmul(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    return [mx.matmul(inputs[0], inputs[1])]

@register("Gemm")
def gemm(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    A, B = inputs[0], inputs[1]
    C = inputs[2] if len(inputs) > 2 and inputs[2] is not None else None
    alpha = attrs.get("alpha", 1.0)
    beta = attrs.get("beta", 1.0)
    transA = attrs.get("transA", 0)
    transB = attrs.get("transB", 0)
    if transA:
        A = mx.transpose(A)
    if transB:
        B = mx.transpose(B)
    result = mx.matmul(A, B)
    if alpha != 1.0:
        result = mx.multiply(result, alpha)
    if C is not None:
        if beta != 1.0:
            C = mx.multiply(C, beta)
        result = mx.add(result, C)
    return [result]

@register("Einsum")
def einsum(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    equation = attrs.get("equation", "")
    return [mx.einsum(equation, *inputs)]


#
#   onnx2mlx
#   Copyright © 2026 NatML Inc. All Rights Reserved.
#

import mlx.core as mx

from ..context import ConvertContext
from . import register

@register("Add")
def add(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    return [mx.add(inputs[0], inputs[1])]

@register("Sub")
def sub(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    return [mx.subtract(inputs[0], inputs[1])]

@register("Mul")
def mul(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    return [mx.multiply(inputs[0], inputs[1])]

@register("Div")
def div(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    a, b = inputs[0], inputs[1]
    if (
        mx.issubdtype(a.dtype, mx.integer) and
        mx.issubdtype(b.dtype, mx.integer)
    ):
        return [mx.floor_divide(a, b)]
    return [mx.divide(a, b)]

@register("Pow")
def pow_(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    return [mx.power(inputs[0], inputs[1])]

@register("Sqrt")
def sqrt(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    return [mx.sqrt(inputs[0])]

@register("Exp")
def exp(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    return [mx.exp(inputs[0])]

@register("Log")
def log(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    return [mx.log(inputs[0])]

@register("Neg")
def neg(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    return [mx.negative(inputs[0])]

@register("Abs")
def abs_(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    return [mx.abs(inputs[0])]

@register("Reciprocal")
def reciprocal(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    return [mx.reciprocal(inputs[0])]

@register("Floor")
def floor(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    return [mx.floor(inputs[0])]

@register("Ceil")
def ceil(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    return [mx.ceil(inputs[0])]

@register("Sign")
def sign(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    return [mx.sign(inputs[0])]

@register("Mod")
def mod(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    return [mx.remainder(inputs[0], inputs[1])]

@register("Sum")
def sum_(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    result = inputs[0]
    for inp in inputs[1:]:
        if inp is not None:
            result = mx.add(result, inp)
    return [result]

@register("Min")
def min_(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    result = inputs[0]
    for inp in inputs[1:]:
        if inp is not None:
            result = mx.minimum(result, inp)
    return [result]

@register("Max")
def max_(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    result = inputs[0]
    for inp in inputs[1:]:
        if inp is not None:
            result = mx.maximum(result, inp)
    return [result]

@register("Erf")
def erf(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    return [mx.erf(inputs[0])]

@register("Sin")
def sin(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    return [mx.sin(inputs[0])]

@register("Cos")
def cos(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    return [mx.cos(inputs[0])]

@register("Tan")
def tan(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    return [mx.tan(inputs[0])]

@register("Atan")
def atan(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    return [mx.arctan(inputs[0])]

@register("Round")
def round_(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    return [mx.round(inputs[0])]


#
#   onnx2mlx
#   Copyright © 2026 NatML Inc. All Rights Reserved.
#

import mlx.core as mx
from . import register

@register("DynamicQuantizeLinear")
def dynamic_quantize_linear(inputs, attrs, ctx=None) -> list[mx.array]:
    x = inputs[0]
    x_min = mx.minimum(mx.array(0, dtype=x.dtype), mx.min(x))
    x_max = mx.maximum(mx.array(0, dtype=x.dtype), mx.max(x))
    scale = (x_max - x_min) / mx.array(255.0, dtype=x.dtype)
    scale = mx.maximum(scale, mx.array(1e-8, dtype=x.dtype))
    zero_point = mx.clip(mx.round(-x_min / scale), 0, 255).astype(mx.uint8)
    y = mx.clip(mx.round(x / scale) + zero_point.astype(x.dtype), 0, 255).astype(mx.uint8)
    return [y, scale, zero_point]

@register("MatMulInteger")
def matmul_integer(inputs, attrs, ctx=None) -> list[mx.array]:
    a = inputs[0]
    b = inputs[1]
    a_zp = inputs[2] if len(inputs) > 2 and inputs[2] is not None else mx.array(0, dtype=a.dtype)
    b_zp = inputs[3] if len(inputs) > 3 and inputs[3] is not None else mx.array(0, dtype=b.dtype)
    a_f = a.astype(mx.float32) - a_zp.astype(mx.float32)
    b_f = b.astype(mx.float32) - b_zp.astype(mx.float32)
    return [(a_f @ b_f).astype(mx.int32)]

@register("ConvInteger")
def conv_integer(inputs, attrs, ctx=None) -> list[mx.array]:
    x = inputs[0]
    w = inputs[1]
    x_zp = inputs[2] if len(inputs) > 2 and inputs[2] is not None else mx.array(0, dtype=x.dtype)
    w_zp = inputs[3] if len(inputs) > 3 and inputs[3] is not None else mx.array(0, dtype=w.dtype)
    x_f = x.astype(mx.float32) - x_zp.astype(mx.float32)
    w_f = w.astype(mx.float32) - w_zp.astype(mx.float32)
    # Delegate to the Conv handler via NCHW->NHWC conv
    from .conv import _conv_impl
    result = _conv_impl(x_f, w_f, None, attrs)
    return [result.astype(mx.int32)]

@register("DynamicQuantizeLSTM")
def dynamic_quantize_lstm(inputs, attrs, ctx=None) -> list[mx.array]:
    x = inputs[0]                # (seq_len, batch, input_size)
    w_quant = inputs[1]          # (num_directions, input_size, 4*hidden) uint8
    r_quant = inputs[2]          # (num_directions, hidden, 4*hidden) uint8
    b = inputs[3]                # (num_directions, 8*hidden)
    seq_lens = inputs[4]         # (batch,) or None
    init_h = inputs[5]           # (num_directions, batch, hidden)
    init_c = inputs[6]           # (num_directions, batch, hidden)
    # input[7] is empty (peepholes, unused)
    w_scale = inputs[8]          # (num_directions,)
    w_zp = inputs[9]             # (num_directions,) uint8
    r_scale = inputs[10]         # (num_directions,)
    r_zp = inputs[11]            # (num_directions,) uint8
    hidden_size = int(attrs.get("hidden_size"))
    direction = attrs.get("direction", "forward")
    num_directions = 2 if direction == "bidirectional" else 1
    seq_len = x.shape[0]
    batch = x.shape[1]
    all_h_states = []
    final_h = []
    final_c = []
    for d in range(num_directions):
        # Pre-compute integer weight matrices (zero-point subtracted)
        w_int = w_quant[d].astype(mx.float32) - w_zp[d].astype(mx.float32)
        r_int = r_quant[d].astype(mx.float32) - r_zp[d].astype(mx.float32)
        b_d = b[d] if b is not None else mx.zeros((8 * hidden_size,))
        bias = b_d[:4 * hidden_size] + b_d[4 * hidden_size:]
        h_t = init_h[d] if init_h is not None else mx.zeros((batch, hidden_size))
        c_t = init_c[d] if init_c is not None else mx.zeros((batch, hidden_size))
        h_seq = []
        indices = range(seq_len) if d == 0 else range(seq_len - 1, -1, -1)
        for t in indices:
            # Quantize x_t and h_t to uint8, then integer matmul + dequantize
            x_scale, x_int = _dynamic_quantize(x[t])
            h_scale, h_int = _dynamic_quantize(h_t)
            gates = (
                (x_int @ w_int) * (x_scale * w_scale[d])
                + (h_int @ r_int) * (h_scale * r_scale[d])
                + bias
            )
            i = mx.sigmoid(gates[:, :hidden_size])
            o = mx.sigmoid(gates[:, hidden_size:2 * hidden_size])
            f = mx.sigmoid(gates[:, 2 * hidden_size:3 * hidden_size])
            c_cand = mx.tanh(gates[:, 3 * hidden_size:])
            c_t = f * c_t + i * c_cand
            h_t = o * mx.tanh(c_t)
            h_seq.append(h_t)
        if d == 1:
            h_seq.reverse()
        all_h_states.append(mx.stack(h_seq))
        final_h.append(h_t)
        final_c.append(c_t)
    y = mx.stack(all_h_states, axis=1)
    y_h = mx.stack(final_h, axis=0)
    y_c = mx.stack(final_c, axis=0)
    return [y, y_h, y_c]

def _dynamic_quantize(x):
    """
    Quantize float tensor to uint8, returning (scale, zero_point, x_int).

    x_int contains the zero-point-subtracted integer values in float32,
    ready for matmul. This matches the DynamicQuantizeLinear + subtract
    pipeline that ORT uses internally.
    """
    x_min = mx.minimum(mx.array(0.0), mx.min(x))
    x_max = mx.maximum(mx.array(0.0), mx.max(x))
    scale = mx.maximum((x_max - x_min) / 255.0, mx.array(1e-8))
    zp = mx.clip(mx.round(-x_min / scale), 0, 255)
    x_q = mx.clip(mx.round(x / scale) + zp, 0, 255)
    return scale, x_q - zp


#
#   onnx2mlx
#   Copyright © 2026 NatML Inc. All Rights Reserved.
#

import mlx.core as mx

from ..context import ConvertContext
from . import register

@register("Equal")
def equal(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    return [mx.equal(inputs[0], inputs[1])]

@register("Greater")
def greater(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    return [mx.greater(inputs[0], inputs[1])]

@register("Less")
def less(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    return [mx.less(inputs[0], inputs[1])]

@register("GreaterOrEqual")
def greater_or_equal(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    return [mx.greater_equal(inputs[0], inputs[1])]

@register("LessOrEqual")
def less_or_equal(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    return [mx.less_equal(inputs[0], inputs[1])]

@register("Not")
def not_(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    return [mx.logical_not(inputs[0])]

@register("And")
def and_(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    return [mx.logical_and(inputs[0], inputs[1])]

@register("Or")
def or_(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    return [mx.logical_or(inputs[0], inputs[1])]

@register("Where")
def where(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    return [mx.where(inputs[0], inputs[1], inputs[2])]

@register("IsNaN")
def isnan(
    inputs: list[mx.array | None],
    attrs: dict[str, object],
    ctx: ConvertContext,
) -> list[mx.array]:
    return [mx.isnan(inputs[0])]


#
#   onnx2mlx
#   Copyright © 2026 NatML Inc. All Rights Reserved.
#

from __future__ import annotations
import mlx.core as mx
from typing import Callable, TYPE_CHECKING

from ..context import ConvertContext

Handler = Callable[[
    list[mx.array | None],
    dict[str, object],
    ConvertContext
], list[mx.array]]

OP_REGISTRY: dict[str, Handler] = {}

def register(*op_types: str):
    """
    Register one or more ONNX op types to a handler function.
    """
    def decorator(fn: Handler) -> Handler:
        for op_type in op_types:
            OP_REGISTRY[op_type] = fn
        return fn
    return decorator

from . import (
    arithmetic,
    activation,
    comparison,
    linalg,
    conv,
    pool,
    normalization,
    shape,
    reduction,
    tensor,
    image,
    quant,
    sequence,
    control,
)